@@ -6,273 +6,180 @@ import pandas as pd
 import os
 from glob import glob
 from datetime import datetime
+import typing
+from typing import Union, Mapping, Sequence, Tuple, Set, List
 
 # operators like and_, or_, and not_, functions like sum, min, max, etc
-import sqlalchemy.sql as op
-from sqlalchemy.sql import func
-import sqlalchemy as sa
+#import sqlalchemy as sa
+import sqlalchemy
 
-from .coltypes import CpickleType, ParseTreeType, PickleFileType, TextFileType, FileTypeBase, JSONType
+from .coltypes import FileTypeBase
 from .bootstrap import DocBootstrap
-from .util import list_tables
+#from .util import list_tables
+from .connectengine import ConnectEngine
+from .schemas import parse_schema
+
 
 class DocTable:
-    _type_map = {
-        'biginteger':sa.BigInteger,
-        'boolean':sa.Boolean,
-        'date':sa.Date,
-        'datetime':sa.DateTime,
-        'enum':sa.Enum,
-        'float':sa.Float,
-        'integer':sa.Integer,
-        'interval':sa.Interval,
-        'largebinary':sa.LargeBinary,
-        'numeric':sa.Numeric,
-        #'pickle':sa.PickleType,
-        'smallinteger':sa.SmallInteger,
-        'string':sa.String,
-        'text':sa.Text,
-        'time':sa.Time,
-        'unicode':sa.Unicode,
-        'unicodetext':sa.UnicodeText,
-        'json': JSONType, # custom datatype
-        'pickle': CpickleType, # custom datatype
-        'parsetree': ParseTreeType, # custom datatype
-        'picklefile': PickleFileType,
-        'textfile': TextFileType,
-    }
-    
+    ''' Class for managing a single database table.
+    Description: This class manages schema and connection information to provide
+        an object-based interface to perform queries on a single table 
+        in a database (although multi-table designs are possible with 
+        multiple DocTables). It is designed to maintain informabout about the 
+        underlying database structure of the table, making it possible to 
+        execute queries without using the SQL language.
 
-    #_valid_types = list(_constraint_map.keys()) + list(_type_map.keys())
-    
-    def __init__(self, schema=None, tabname='_documents_', fname=':memory:', engine='sqlite', persistent_conn=True, verbose=False, make_new_db=True, **engine_args):
+    Settable static attributes (overridden if related constructor argument passed):
+        __tabname__ (str): name of table to connect to (and create).
+        __schema__ (str): schema definition for this doctable, to be used
+            when creating a new table and to manage table information.
+        __target__ (str): target database to connect to - used when a doctable
+            will always connect to the same target (i.e., a server etc).
+        __args__ (dict): any other constructor arguments that should always
+            be used when instantiating. Overridden by providing arguments
+            to the constructor.
+    '''
+    __default_tabname__ = '_documents_'
+    def __init__(self, target: str = None, tabname: str = None, schema: Sequence[Sequence] = None, dialect='sqlite', engine=None,  
+                 readonly=False, new_db=False, new_table=True, persistent_conn=True, 
+                 verbose=False, **engine_kwargs):
         '''Create new database.
         Args:
-            schema (list<list>): schema from which to create db. Includes a
-                list of columns (including contraints and indexes) as tuples
-                defined according to information needed to execute the sqlalchemy
-                commands.
-            tabname (str): table name for this specific doctable.
-            fname (str): filename for database to connect to. ":memory:" is a 
-                special value indicating to the python db engine that the db
+            target (str): filename for database to connect to. ":memory:" is a 
+                special value indicating to the python sqlite engine that the db
                 should be created in memory. Will create new empty database file
-                if it does not exist and new_db is True.
-            engine (str): database engine through which to construct db.
+                if it does not exist and new_db==True, and add a new table using
+                specified schema if new_table==True.
+            schema (list<list>): schema from which to create db. Includes a
+                list of column names and types (including contraints and indexes) as tuples
+                defined according to information needed to construct the sqlalchemy
+                objects.
+            tabname (str): table name for this specific doctable.
+            dialect (str): database engine through which to construct db.
                 For more info, see sqlalchemy dialect info:
                 https://docs.sqlalchemy.org/en/13/dialects/
             persistent_conn (bool): whether or not to create a persistent conn 
-                to database. Set to True to lock db from other process access 
-                while instance exists, esp if calling .update() in a .select()
-                loop. Set to False to access from separate processes.
-            verbose (bool): Print every sql command before executing.
+                to database. Otherwise will create temporary connection for each
+                query.
+            readonly (bool): Prevents user from calling insert(), delete(), or 
+                update(). Will not block other sql possible commands.
             new_db (bool): Indicate if new db file should be created given 
                 that a schema is provided and the db file doesn't exist.
-            engine_args (**kwargs): Pass directly to the sqlalchemy
+            new_table (bool): Allow doctable to create a new table if one 
+                doesn't exist already.
+            engine_kwargs (**kwargs): Pass directly to the sqlalchemy
                 .create_engine(). Args typically vary by dialect.
                 Example: connect_args={'timeout': 15} for sqlite
                 or connect_args={'connect_timeout': 15} for PostgreSQL.
+            verbose (bool): Print every sql command before executing.
+            echo (bool): Print sqlalchemy engine log for each query.
         '''
         
-        # in cases where user did not want to create new db but a db does not 
-        # exist
+        # look for statically defined class members
+        try:
+            self.__tabname__
+            if tabname is None:
+                tabname = self.__tabname__
+        except AttributeError:
+            pass
+        try:
+            self.__schema__
+            if schema is None:
+                schema = self.__schema__
+        except AttributeError:
+            pass
+        try:
+            self.__target__
+            if target is None:
+                target = self.__target__
+        except AttributeError:
+            pass
         
-        if fname != ':memory:' and not os.path.exists(fname) and not make_new_db:
-            raise FileNotFoundError('make_new_db is set to False but the database does not '
-                             'exist yet.')
-        elif schema is None and (fname == ':memory:' or not os.path.exists(fname)):
-            raise ValueError('Schema must be provided if using memory database or '
+        
+        # check for argument information in static member variable "args"
+        try:
+            kwargs = self.__args__ # user can provide their own data
+            
+            # these will definitely override constructor args
+            if 'dialect' in kwargs:
+                dialect = kwargs['dialect']
+            if 'verbose' in kwargs:
+                verbose = kwargs['verbose']
+            if 'new_db' in kwargs:
+                new_db = kwargs['new_db']
+            if 'engine_kwargs' in kwargs:
+                engine_kwargs = kwargs['engine_kwargs']
+            
+            # these will override if not provided
+            if tabname is None and 'tabname' in kwargs:
+                tabname = kwargs['tabname']
+            if schema is None and 'schema' in kwargs:
+                schema = kwargs['schema']
+            if target is None and 'target' in kwargs:
+                target = kwargs['target']
+            
+        except AttributeError:
+            pass
+        
+        # set defaults
+        if tabname is None:
+            tabname = self.__default_tabname__
+        if engine is not None:
+            target = engine.target
+        if readonly:
+            new_db = False
+            new_table = False
+            
+        # run some checks
+        if target is None:
+            raise ValueError('target has not been provided.')
+        
+        if dialect.startswith('sqlite'):
+            if schema is None and (target == ':memory:' or not os.path.exists(target)):
+                raise ValueError('Schema must be provided if using memory database or '
                              'database file does not exist yet. Need to provide schema '
                              'when creating a new table.')
-        elif schema is None and tabname not in list_tables(fname, engine=engine, **engine_args):
-            tables = list_tables(fname, engine=engine, **engine_args)
-            raise ValueError('The requested table was not found in the database! Schema must be '
-                             'provided to create a new table. Existing tables: {tables}')
         
-        # separate tables for custom data types and main table
-        self._fname = fname
+        # store arguments as-is
         self._tabname = tabname
+        self._target = target
         self.verbose = verbose
+        self.user_schema = schema
+        self.persistent_conn = persistent_conn
+        self._readonly = readonly
+        self._new_db = new_db
+        self._new_table = new_table
         
-        # create database engine
-        self._schema = schema
-        self._connstr = '{}:///{}'.format(engine,fname)
-        self._engine_args = engine_args
-        self._conn = None
-        self.open_engine(open_conn=persistent_conn)
-        
-        # bind .min(), .max(), and .count() to col objects themselves.
-        self._bind_functions()
-
-    
-    def __delete__(self):
-        '''Closes database connection to prevent locking.'''
-        self.close_conn()
-            
-    def close_engine(self):
-        self.close_conn()
-        self._engine = None
-    
-    def open_engine(self, open_conn=True):
-        '''Open engine manually (esp when concurrency applications).'''
-        
-        # make new engine
-        self._engine = sa.create_engine(self._connstr, **self._engine_args)
-    
-        # make table if needed
-        self._metadata = sa.MetaData()
-        if self._schema is not None:
-            columns = self._parse_column_schema(self._schema)
-            self._table = sa.Table(self._tabname, self._metadata, *columns)
-            self._metadata.create_all(self._engine)
+        # establish an engine connection
+        if engine is None:
+            self._engine = ConnectEngine(target=target, dialect=dialect, new_db=new_db, 
+                                     **engine_kwargs)
         else:
-            # NOTE: use list_tables to list available tables upon failure!
-            self._table = sa.Table(self._tabname, self._metadata, 
-                                   autoload=True, autoload_with=self._engine)
-        if open_conn:
-            self.open_conn()
-            
-            
-    def close_conn(self):
-        '''Closes connection to db (if one exists).
-        Notes:
-            Primarily to be used if persistent_conn flag was set
-                to true in constructor, but user wants to close.
+            self._engine = engine
+        
+        # connect to existing table or create new one
+        self._columns = None
+        if schema is not None:
+            self._columns = parse_schema(schema, target+'_'+tabname)
+        
+        self._table = self._engine.add_table(self._tabname, columns=self._columns, 
+                                             new_table=self._new_table)
+        
+        # connect to database
+        self._conn = self._engine.get_connection()
+        
+    def __del__(self):
+        ''' Closes database connection to prevent locking db.
         '''
-        if self._conn is not None:
-            self._conn.close()
-        self._conn = None
-        
-    def open_conn(self):
-        '''Opens connection to db (if one does not exist).
-        Notes:
-            Primarily to be used if persistent_conn flag was set
-                to false in constructor, but user wants to create.
-        
-        '''
-        if self._engine is None:
-            raise ValueError('Need to create engine using .open_engine() '
-                             'before trying to open connection.')
-        
-        if self._conn is None:
-            self._conn = self._engine.connect()
-        
-        
-    ################# INITIALIZATION METHODS ##################
-    
-    def _parse_column_schema(self,schema):
-        columns = list()
-        for colinfo in schema:
-            n = len(colinfo)
-            if n not in (2,3,4):
-                raise ValueError('A schema entry must have 2+ arguments: (type,name,..)')
-
-            # column is regular type
-            if colinfo[0] in self._type_map:
-                coltype, colname = colinfo[:2]
-                colargs = colinfo[2] if n > 2 else dict()
-                coltypeargs = colinfo[3] if n > 3 else dict()
-                if coltype in ('picklefile','textfile') and self._fname != ':memory:':
-                    if 'fpath' not in coltypeargs:
-                        #path = os.path.splitext(self._fname)[0]
-                        #coltypeargs['fpath'] = path+'_'+self._tabname+'_'+colname
-                        coltypeargs['fpath'] = self._fname+'_'+self._tabname+'_'+colname
-                
-                col = sa.Column(colname, self._type_map[coltype](**coltypeargs), **colargs)
-                columns.append(col)
-            else:
-                if colinfo[0] == 'idcol': #shortcut for typical id integer primary key etc
-                    col = sa.Column(colinfo[1], sa.Integer, primary_key=True, autoincrement=True)
-                    columns.append(col)
-                elif colinfo[0] == 'date_added': #shortcut for typical id integer primary key etc
-                    col = sa.Column(colinfo[1], sa.DateTime, default=datetime.now)
-                    columns.append(col)
-                elif colinfo[0] == 'date_updated': #shortcut for typical id integer primary key etc
-                    col = sa.Column(colinfo[1], sa.DateTime, default=datetime.now, onupdate=datetime.now)
-                    columns.append(col)
-                elif colinfo[0] == 'index':
-                    # ('index', 'ind0', ('name','address'),dict(unique=True)),
-                    indargs = colinfo[2] if n > 2 else dict()
-                    indkwargs = colinfo[3] if n > 3 else dict()
-                    ind = sa.Index(colinfo[1], *indargs, **indkwargs)
-                    columns.append(ind)
-                elif colinfo[0] == 'check_constraint':
-                    # ('check_constraint','age >= 0 AND age < 120',, dict(name='age')),
-                    kwargs = colinfo[2] if n > 2 else dict()
-                    const = sa.CheckConstraint(colinfo[1], **kwargs)
-                    columns.append(const)
-                elif colinfo[0] == 'unique_constraint':
-                    # ('unique_constraint', ('name','address'), dict(name='name_addr')),
-                    kwargs = colinfo[2] if n > 2 else dict()
-                    const = sa.UniqueConstraint(*colinfo[1], **kwargs)
-                    columns.append(const)
-                elif colinfo[0] == 'primarykey_constraint':
-                    # ('primarykey_constraint', ('name','address'),dict(unique=True)),
-                    kwargs = colinfo[2] if n > 2 else dict()
-                    const = sa.PrimaryKeyConstraint(*colinfo[1], **kwargs)
-                    columns.append(const)
-                elif colinfo[0] == 'foreignkey_constraint':
-                    # ('foreignkey_constraint', 
-                    # [('othertab_name','othertab_address'),('othertab.name', 'othertab.address')])
-                    kwargs = colinfo[2] if n > 2 else dict()
-                    const = sa.ForeignKeyConstraint(*colinfo[1], **kwargs)
-                    columns.append(const)
-                else:
-                    raise ValueError('Column or constraint type "{}" was not recognized.'
-                                    ''.format(colinfo[0]))
-                
-        return columns
-
-    
-    def _bind_functions(self):
-        '''Binds .max(), .min(), .count() to each column object.
-            note: https://docs.sqlalchemy.org/en/13/core/functions.html
-        '''
-        for col in self._table.c:
-            col.max = func.max(col)
-            col.min = func.min(col)
-            col.count = func.count(col)
-            col.sum = func.sum(col)
-            
-    def clean_col_files(self, col, check_missing=True, delete_extraneous=True):
-        '''Make sure there is a 1-1 mapping between files listed in db and files in folder.
-        Args:
-            col (str or Column object): column to clean picklefiles for.
-            ignore_missing (bool): if False, throw an error when a db file doesn't exist.
-        '''
-        col = self.col(col)
-        if not isinstance(self.col(col).type, FileTypeBase):
-            raise ValueError('Only call clean_col_files for a file column types.')
-        if not (check_missing or delete_extraneous):
-            raise ValueError('Either check_missing or delete_extraneous should be set to true, '
-                             'or else this method does nothing.')
-        
-        # get column filenames
-        db = DocTable(fname=self._fname, tabname=self._tabname)
-        dbcol = db[col.name]
-        db_fnames = {col.type.fpath+fn for fn in db.select(dbcol, where=dbcol != None)}
-        
-        # get existing files from filesystem
-        exist_fnames = set(glob(col.type.fpath+'*'+col.type.file_ext))
-        intersect = db_fnames & exist_fnames
-        
-        # remove files not listed in db
-        extraneous_fnames = exist_fnames - intersect
-        if delete_extraneous and len(extraneous_fnames) > 0:
-            for rm_fname in extraneous_fnames:
-                os.remove(rm_fname)
-        
-        # throw error if filesystem is missing some files
-        miss_fnames = db_fnames-intersect
-        if check_missing and len(miss_fnames) > 0:
-            raise FileNotFoundError('These files were not found while cleaning: {}'
-                ''.format(miss_fnames))
-                
+        #self._engine.remove_table(self._table) # remove from engine metadata
+        self.close_conn()
     
     #################### Convenience Methods ###################
     
-    def __str__(self):
-        return '<DocTable::{} ct: {}>'.format(self._tabname, self.count())
-    def __repr__(self):
+    def __str__(self) -> str:
+        return '<DocTable::{}:{} ct: {}>'.format(repr(self._engine), self._tabname, self.count())
+    
+    def __repr__(self) -> str:
         return str(self)
     
     def __getitem__(self, colname):
@@ -280,12 +187,12 @@ class DocTable:
         return self.col(colname)
     
     def col(self,name):
-        '''Accesses a column object. Equivalent to table.c[name].
+        '''Accesses a column object.
         Args:
             Name of column to access. Applied as subscript to 
                 sqlalchemy columns object.
         '''
-        if isinstance(name, sa.Column):
+        if isinstance(name, sqlalchemy.Column):
             return name
         return self._table.c[name]
         
@@ -316,18 +223,53 @@ class DocTable:
         return self._table.c
     
     @property
+    def engine(self):
+        return self._engine
+    
+    def list_tables(self):
+        return self._engine.list_tables()
+    
     def colnames(self):
         return [c.name for c in self.columns]
     
-    @property
-    def schemainfo(self):
+    def schema_info(self):
         '''Get info about each column as a dictionary.
         Returns:
             dict<dict>: info about each column.
         '''
-        inspector = sa.inspect(self._engine)
-        return inspector.get_columns(self._tabname)
+        return self._engine.schema()
     
+    def schema_table(self):
+        '''Get info about each column as a dictionary.
+        Returns:
+            DataFrame: info about each column.
+        '''
+        return self._engine.schema_df(self._tabname)
+    
+            
+    #################### Connection Methods ###################
+    
+    def close_conn(self):
+        ''' Closes connection to db (if one exists). '''
+        if self._conn is not None:
+            self._conn.close()
+            self._conn = None
+        
+    def open_conn(self):
+        ''' Opens connection to db (if one does not exist). '''
+        
+        if self._conn is None:
+            self._conn = self._engine.get_connection()
+    
+    def reopen_engine(self, open_conn=None):
+        ''' Opens connection engine. 
+        Args:
+            open_conn (bool): create a new db connection.
+        '''
+        self.close_conn()
+        self._engine.reopen()
+        if open_conn or (open_conn is None and self.persistent_conn):
+            self.open_conn()
     
     ################# INSERT METHODS ##################
     
@@ -340,7 +282,10 @@ class DocTable:
         Returns:
             sqlalchemy query result object. 
         '''
-        q = sa.sql.insert(self._table, rowdat)
+        if self._readonly:
+            raise ValueError('Cannot call .insert() when doctable set to readonly.')
+        
+        q = sqlalchemy.sql.insert(self._table, rowdat)
         q = q.prefix_with('OR {}'.format(ifnotunique.upper()))
         
         #NOTE: there is a weird issue with using verbose mode with a 
@@ -364,18 +309,18 @@ class DocTable:
     ################# SELECT METHODS ##################
     
     
-    def count(self, where=None, whrstr=None, **kwargs):
+    def count(self, where=None, wherestr=None, **kwargs):
         '''Count number of rows which match where condition.
         Notes:
             Calls select_first under the hood.
         Args:
             where (sqlalchemy condition): filter rows before counting.
-            whrstr (str): filter rows before counting.
+            wherestr (str): filter rows before counting.
         Returns:
-            int: number of rows that match "where" and "whrstr" criteria.
+            int: number of rows that match "where" and "wherestr" criteria.
         '''
-        cter = func.count(self._table)
-        ct = self.select_first(cter, where=where, whrstr=whrstr, **kwargs)
+        cter = sqlalchemy.sql.func.count(self._table)
+        ct = self.select_first(cter, where=where, wherestr=wherestr, **kwargs)
         return ct
     
     def head(self, n=5):
@@ -387,7 +332,7 @@ class DocTable:
         '''
         return self.select_df(limit=n)
     
-    def select(self, cols=None, where=None, orderby=None, groupby=None, limit=None, whrstr=None, offset=None, **kwargs):
+    def select(self, cols=None, where=None, orderby=None, groupby=None, limit=None, wherestr=None, offset=None, **kwargs):
         '''Perform select query, yield result for each row.
         
         Description: Because output must be iterable, returns special column results 
@@ -400,7 +345,7 @@ class DocTable:
             orderby: sqlalchemy orderby directive
             groupby: sqlalchemy gropuby directive
             limit (int): number of entries to return before stopping
-            whrstr (str): raw sql "where" conditionals to add to where input
+            wherestr (str): raw sql "where" conditionals to add to where input
         Yields:
             sqlalchemy result object: row data
         '''
@@ -414,7 +359,7 @@ class DocTable:
         cols = [c if not isinstance(c,str) else self[c] for c in cols]
                 
         # query colunmns in main table
-        result = self._exec_select_query(cols,where,orderby,groupby,limit,whrstr,offset,**kwargs)
+        result = self._exec_select_query(cols,where,orderby,groupby,limit,wherestr,offset,**kwargs)
         # this is the result object:
         # https://kite.com/python/docs/sqlalchemy.engine.ResultProxy
         
@@ -488,14 +433,14 @@ class DocTable:
         sel = self.select(col, *args, **kwargs)
         return pd.Series(sel)
     
-    def _exec_select_query(self, cols, where, orderby, groupby, limit, whrstr, offset,**kwargs):
+    def _exec_select_query(self, cols, where, orderby, groupby, limit, wherestr, offset,**kwargs):
         
-        q = sa.sql.select(cols)
+        q = sqlalchemy.sql.select(cols)
         
         if where is not None:
             q = q.where(where)
-        if whrstr is not None:
-            q = q.where(sa.text(whrstr))
+        if wherestr is not None:
+            q = q.where(sqlalchemy.text(wherestr))
         if orderby is not None:
             if is_sequence(orderby):
                 q = q.order_by(*orderby)
@@ -568,7 +513,7 @@ class DocTable:
     
     #################### Update Methods ###################
     
-    def update(self, values, where=None, whrstr=None, **kwargs):
+    def update(self, values, where=None, wherestr=None, **kwargs):
         '''Update row(s) assigning the provided values.
         Args:
             values (dict<colname->value> or list<dict> or list<(col,value)>)): 
@@ -583,23 +528,25 @@ class DocTable:
                 undefined.
             where (sqlalchemy condition): used to match rows where
                 update will be applied.
-            whrstr (sql string condition): matches same as where arg.
+            wherestr (sql string condition): matches same as where arg.
         Returns:
             SQLAlchemy result proxy object
         '''
+        if self._readonly:
+            raise ValueError('Cannot call .update() when doctable set to readonly.')
             
         # update the main column values
         if isinstance(values,list) or isinstance(values,tuple):
-            q = sa.sql.update(self._table, preserve_parameter_order=True)
+            q = sqlalchemy.sql.update(self._table, preserve_parameter_order=True)
             q = q.values(values)
         else:
-            q = sa.sql.update(self._table)
+            q = sqlalchemy.sql.update(self._table)
             q = q.values(values)
         
         if where is not None:
             q = q.where(where)
-        if whrstr is not None:
-            q = q.where(sa.text(whrstr))
+        if wherestr is not None:
+            q = q.where(sqlalchemy.text(wherestr))
         
         r = self.execute(q, **kwargs)
         
@@ -609,23 +556,26 @@ class DocTable:
     
     #################### Delete Methods ###################
     
-    def delete(self, where=None, whrstr=None, vacuum=False, **kwargs):
+    def delete(self, where=None, wherestr=None, vacuum=False, **kwargs):
         '''Delete rows from the table that meet the where criteria.
         Args:
             where (sqlalchemy condition): criteria for deletion.
-            whrstr (sql string): addtnl criteria for deletion.
+            wherestr (sql string): addtnl criteria for deletion.
             vacuum (bool): will execute vacuum sql command to reduce
                 storage space needed by SQL table. Use when deleting
                 significant ammounts of data.
         Returns:
             SQLAlchemy result proxy object.
         '''
-        q = sa.sql.delete(self._table)
+        if self._readonly:
+            raise ValueError('Cannot call .delete() when doctable set to readonly.')
+        
+        q = sqlalchemy.sql.delete(self._table)
 
         if where is not None:
             q = q.where(where)
-        if whrstr is not None:
-            q = q.where(sa.text(whrstr))
+        if wherestr is not None:
+            q = q.where(sqlalchemy.text(wherestr))
         
         r = self.execute(q, **kwargs)
         
@@ -646,7 +596,7 @@ class DocTable:
             verbose (bool or None): Print SQL command issued before
                 execution.
         '''
-        prstr = 'DocTable2 Query: {}'
+        prstr = 'DocTable: {}'
         if verbose is not None:
             if verbose: print(prstr.format(query))
         elif self.verbose: print(prstr.format(query))
@@ -656,10 +606,6 @@ class DocTable:
         return result
     
     def _execute(self, query, conn=None):
-        if self._engine is None:
-            raise AttributeError('DocTable does not have an engine '
-                'connection. Use .open_engine() to make new engine '
-                'connection.')
         
         # takes raw query object
         if conn is not None:
@@ -667,7 +613,7 @@ class DocTable:
         elif self._conn is not None:
             r = self._conn.execute(query)
         else:
-            with self._engine.connect() as conn:
+            with self._engine.get_connection() as conn:
                 r = conn.execute(query)
         return r
     
@@ -690,8 +636,49 @@ class DocTable:
         docs = self.select(*args, **kwargs)
         return DocBootstrap(docs, n=n)
     
+    
+    
+    
+    ################# FILE COLUMN METHODS ##################
 
+    def clean_col_files(self, col, check_missing=True, delete_extraneous=True):
+        '''Make sure there is a 1-1 mapping between files listed in db and files in folder.
+        Args:
+            col (str or Column object): column to clean picklefiles for.
+            ignore_missing (bool): if False, throw an error when a db file doesn't exist.
+        '''
+        col = self.col(col)
+        if not isinstance(self.col(col).type, FileTypeBase):
+            raise ValueError('Only call clean_col_files for a file column types.')
+        if not (check_missing or delete_extraneous):
+            raise ValueError('Either check_missing or delete_extraneous should be set to true, '
+                             'or else this method does nothing.')
+        
+        # get column filenames
+        db = DocTable(tabname=self._tabname, engine=self._engine)
+        dbcol = db[col.name]
+        db_fnames = {col.type.fpath+fn for fn in db.select(dbcol, where=dbcol != None)}
+        
+        # get existing files from filesystem
+        exist_fnames = set(glob(col.type.fpath+'*'+col.type.file_ext))
+        intersect = db_fnames & exist_fnames
+        
+        # remove files not listed in db
+        extraneous_fnames = exist_fnames - intersect
+        if delete_extraneous and len(extraneous_fnames) > 0:
+            for rm_fname in extraneous_fnames:
+                os.remove(rm_fname)
+        
+        # throw error if filesystem is missing some files
+        miss_fnames = db_fnames-intersect
+        if check_missing and len(miss_fnames) > 0:
+            raise FileNotFoundError('These files were not found while cleaning: {}'
+                ''.format(miss_fnames))
+    
+    
+    
 def is_sequence(obj):
     return isinstance(obj, list) or isinstance(obj,set) or isinstance(obj,tuple)
 
-
+def is_ord_sequence(obj):
+    return isinstance(obj, list) or isinstance(obj,tuple)
